import os


class Mute():
    NULL_FDS = []
    BACKUP_FDS = []

    def __enter__(self):
        self.suppress_stdout()

    def __exit__(self, type, value, traceback):
        self.restore_stdout()

    def suppress_stdout(self):
        self.NULL_FDS = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
        self.BACKUP_FDS = os.dup(1), os.dup(2)
        os.dup2(self.NULL_FDS[0], 1)
        os.dup2(self.NULL_FDS[1], 2)

    def restore_stdout(self):
        os.dup2(self.BACKUP_FDS[0], 1)
        os.dup2(self.BACKUP_FDS[1], 2)
        os.close(self.NULL_FDS[0])
        os.close(self.NULL_FDS[1])



#!/usr/bin/env python
import logging
from airflow import configuration
from cwl_airflow.utils.func import make_dag, get_active_jobs


def create_dags():
    dags = {}
    for job in get_active_jobs(jobs_folder=configuration.get('cwl', 'jobs'), limit=int(configuration.get('cwl', 'limit'))):
        try:
            dags[job["dag_id"]] = make_dag(job)
        except Exception:
            logging.error("Failed to create DAG for \n{}\n{}".format(job["content"]["workflow"], job["path"]))
            pass
    return dags



import os
import sys
import urllib.parse
import logging
import argparse
import re
from json import dumps
import cwltool.context
from cwltool.workflow import default_make_tool
from cwltool.resolver import tool_resolver
import cwltool.load_tool as load
from cwltool.argparser import get_default_args
from cwl_airflow.utils.mute import Mute
from schema_salad.ref_resolver import Loader
from airflow import configuration
from airflow.exceptions import AirflowConfigException
from airflow.models import DagRun


def norm_path(path):
    return os.path.abspath(os.path.normpath(os.path.normcase(path)))


def get_files(current_dir, filename_pattern=".*"):
    """Files with the identical basenames are overwritten"""
    files_dict = {}
    for root, dirs, files in os.walk(current_dir):
        files_dict.update(
            {filename: os.path.join(root, filename) for filename in files if re.match(filename_pattern, filename)}
        )
    return files_dict


def convert_to_workflow(tool, tool_file, workflow_file):
    workflow = {"cwlVersion": "v1.0",
                "class":      "Workflow",
                "inputs":     gen_inputs(tool),
                "outputs":    gen_outputs(tool)}
    workflow["steps"] = [{"run": tool_file,
                          "in": [{"source": inp["id"], "id": inp["id"]} for inp in workflow["inputs"]],
                          "out": [outp["id"] for outp in workflow["outputs"]],
                          "id": "static_step"}]
    if tool.get("$namespaces", None):
        workflow["$namespaces"] = tool["$namespaces"]
    if tool.get("$schemas", None):
        workflow["$schemas"] = tool["$schemas"]
    if tool.get("requirements", None):
        workflow["requirements"] = tool["requirements"]

    with open(workflow_file, 'w') as output_stream:
        output_stream.write(dumps(workflow, indent=4))

    return workflow_file


def gen_inputs(tool):
    inputs = []
    for inpt in tool["inputs"]:
        custom_input = {}
        if inpt.get("id", None):
            custom_input["id"] = shortname(inpt["id"])
        if inpt.get("type", None):
            custom_input["type"] = inpt["type"]
        if inpt.get("format", None):
            custom_input["format"] = inpt["format"]
        if inpt.get("doc", None):
            custom_input["doc"] = inpt["doc"]
        if inpt.get("label", None):
            custom_input["label"] = inpt["label"]
        inputs.append(custom_input)
    return inputs


def gen_outputs(tool):
    outputs = []
    for outp in tool["outputs"]:
        custom_output = {}
        if outp.get("id", None):
            custom_output["id"] = shortname(outp["id"])
        if outp.get("type", None):
            custom_output["type"] = outp["type"]
        if outp.get("format", None):
            custom_output["format"] = outp["format"]
        if outp.get("doc", None):
            custom_output["doc"] = outp["doc"]
        if outp.get("label", None):
            custom_output["label"] = outp["label"]
        if outp.get("id", None):
            custom_output["outputSource"] = "static_step/" + shortname(outp["id"])
        outputs.append(custom_output)
    return outputs


def load_cwl(cwl_file):
    load.loaders = {}
    loading_context = cwltool.context.LoadingContext(get_default_args())
    loading_context.construct_tool_object = default_make_tool
    loading_context.resolver = tool_resolver
    return load.load_tool(cwl_file, loading_context)


def exit_if_unsupported_feature(cwl_file, exit_code=33):
    with Mute():
        tool = load_cwl(cwl_file).tool
    if tool["class"] == "Workflow" and [step["id"] for step in tool["steps"] if "scatter" in step]:
        logging.warning("Failed to submit workflow\n- {} - scatter is not supported".format(cwl_file))
        sys.exit(exit_code)


def normalize_args(args, skip_list=[]):
    """Converts all relative path arguments to absolute ones relatively to the current working directory"""
    normalized_args = {}
    for key, value in args.__dict__.items():
        if key not in skip_list:
            normalized_args[key] = value if not value or os.path.isabs(value) else os.path.normpath(os.path.join(os.getcwd(), value))
        else:
            normalized_args[key]=value
    return argparse.Namespace(**normalized_args)


def shortname(n):
    return n.split("#")[-1]


def load_job(job_file):
    loader = Loader(load.jobloaderctx.copy())
    job_order_object, _ = loader.resolve_ref(job_file, checklinks=False)
    return job_order_object


def open_file(filename):
    """Returns list of lines from the text file. \n at the end of the lines are trimmed. Empty lines are excluded"""
    lines = []
    with open(filename, 'r') as infile:
        for line in infile:
            if line.strip():
                lines.append(line.strip())
    return lines


def export_to_file (output_filename, data):
    with open(output_filename, 'w') as output_file:
        output_file.write(data)


def get_folder(abs_path, permissions=0o0775, exist_ok=True):
    os.makedirs(abs_path, mode=permissions, exist_ok=exist_ok)
    return abs_path


def list_files(abs_path, ext=[None]):
    all_files = []
    for root, dirs, files in os.walk(abs_path):
        all_files.extend([os.path.join(root, filename) for filename in files
                          if not ext or os.path.splitext(filename)[1] in ext])
    return all_files


def flatten(input_list):
    result = []
    for i in input_list:
        if isinstance(i,list): result.extend(flatten(i))
        else: result.append(i)
    return result


def url_shortname(inputid):
    d = urllib.parse.urlparse(inputid)
    if d.fragment:
        return d.fragment.split(u"/")[-1]
    else:
        return d.path.split(u"/")[-1]


def conf_get_default (section, key, default):
    try:
        return configuration.get(section, key)
    except AirflowConfigException:
        return default


def set_permissions (item, dir_perm=0o0777, file_perm=0o0666, grp_own=os.getgid(), user_own=-1):
    os.chown(item, user_own, grp_own)
    if os.path.isfile(item):
        os.chmod(item, file_perm)
    else:
        os.chmod(item, dir_perm)
        for root, dirs, files in os.walk(item):
            for file in files:
                os.chmod(os.path.join(root,file), file_perm)
                os.chown(os.path.join(root,file), user_own, grp_own)
            for directory in dirs:
                os.chmod(os.path.join(root,directory), dir_perm)
                os.chown(os.path.join(root,directory), user_own, grp_own)


def gen_dag_id(job_file):
    job_content = load_job(job_file)
    w_seg = ".".join(job_content["workflow"].split("/")[-1].split(".")[0:-1])
    j_seg = ".".join(job_file.split("/")[-1].split(".")[0:-1])
    return "-".join([w_seg, j_seg, job_content["uid"]])


def eval_log_level(key):
    log_depth = {
        'CRITICAL': 50,
        'ERROR': 40,
        'WARNING': 30,
        'INFO': 20,
        'DEBUG': 10,
        'NOTSET': 0
    }
    return log_depth[key] if key in log_depth else 20


def set_logger():
    cwl_logger = logging.getLogger("cwltool")
    cwl_logger.addHandler(logging.StreamHandler())
    cwl_logger.setLevel(eval_log_level(conf_get_default('cwl','LOG_LEVEL','INFO').upper()))


def get_latest_log(dag_id, task_id="JobCleanup", state=None):
    log_base = os.path.expanduser(configuration.get('core', 'BASE_LOG_FOLDER'))
    dag_run = sorted(DagRun.find(dag_id, state=state), reverse=True, key=lambda x: x.execution_date)[0]
    for task in dag_run.get_task_instances():
        if task.task_id == task_id:
            kwargs = {"log_base": log_base,
                      "dag_id": task.dag_id,
                      "task_id": task.task_id,
                      "execution_date": task.execution_date.isoformat(),
                      "try_number": task._try_number
            }
            return "{log_base}/{dag_id}/{task_id}/{execution_date}/{try_number}.log".format(**kwargs)


def get_workflow_output(dag_id):  # make something more relyable that splitting by Subtask
    logging.info("Workflow results")
    found_output = False
    results = []
    for line in open_file(get_latest_log(dag_id)):
        if 'WORKFLOW RESULTS' in line:
            found_output = True
            continue
        if found_output:
            try:
                results.append(line.split('Subtask: ')[1])
            except Exception:  # Better use IndexError instead of Exception
                pass
    return "\n".join(results)



import sys
import json
import logging
import tempfile
from argparse import Namespace
from airflow.models import BaseOperator
from cwltool.main import init_job_order
from cwltool.load_tool import jobloaderctx
from schema_salad.ref_resolver import Loader
from cwl_airflow.utils.utils import shortname

class JobDispatcher(BaseOperator):

    def __init__(self, *args, **kwargs):
        super(JobDispatcher, self).__init__(task_id=self.__class__.__name__, *args, **kwargs)

    def execute(self, context):
        initialized_job_order_object = init_job_order(self.dag.default_args["job_data"]["content"],
                                                      Namespace(),
                                                      self.dag.cwlwf,
                                                      Loader(jobloaderctx.copy()),
                                                      sys.stdout)

        updated_job_order_object = {}
        for index, inp in enumerate(self.dag.cwlwf.tool["inputs"]):
            inp_id = shortname(inp["id"])
            if inp_id.split("/")[-1] in initialized_job_order_object:
                updated_job_order_object[inp_id] = initialized_job_order_object[inp_id.split("/")[-1]]

        updated_job_order_object["tmp_folder"] = tempfile.mkdtemp(dir=self.dag.default_args["job_data"]["content"].get("tmp_folder", None), prefix="dag_tmp_")
        updated_job_order_object["output_folder"] = self.dag.default_args["job_data"]["content"]["output_folder"]
        logging.info("Dispatch job\n{}".format(json.dumps(updated_job_order_object, indent=4)))
        return {"outputs": updated_job_order_object}



#!/usr/bin/env python
import sys
import os
import argparse
import uuid
import logging
from cwl_airflow.utils.mute import Mute
with Mute():  # Suppress output
    from airflow.settings import AIRFLOW_HOME
    from cwl_airflow.utils.logger import reset_root_logger
    from airflow.bin.cli import scheduler, webserver, initdb
    from cwl_airflow.utils.func import (export_job_file,
                                        add_run_info,
                                        update_config,
                                        export_dags,
                                        create_folders,
                                        get_demo_workflow,
                                        get_updated_args,
                                        start_background_scheduler,
                                        get_airflow_default_args,
                                        clean_jobs_folder,
                                        get_webserver_url,
                                        asset_conf)
    from cwl_airflow.utils.utils import (get_workflow_output,
                                         normalize_args,
                                         exit_if_unsupported_feature,
                                         conf_get_default)


def arg_parser():
    parent_parser = argparse.ArgumentParser(add_help=False)
    general_parser = argparse.ArgumentParser(description='cwl-airflow')

    general_parser.add_argument("-q", "--quiet", dest='quiet', action="store_true", help="Suppress all output except warnings and errors")

    subparsers = general_parser.add_subparsers()
    subparsers.required = True


    init_parser = subparsers.add_parser('init', help="Init cwl-airflow", parents=[parent_parser])
    init_parser.set_defaults(func=run_init)
    init_parser.add_argument("-l", "--limit",    dest='limit', type=int, help="Limit job concurrancy",
                             default=conf_get_default("cwl", "limit", 10))
    init_parser.add_argument("-j", "--jobs",     dest='jobs', type=str, help="Jobs folder. Default: ~/airflow/jobs",
                             default=conf_get_default("cwl", "jobs", os.path.join(AIRFLOW_HOME, 'jobs')))
    init_parser.add_argument("-t", "--timeout",  dest='dag_timeout', type=int, help="How long before timing out a python file import while filling the DagBag",
                             default=conf_get_default("core", "dagbag_import_timeout", 30))
    init_parser.add_argument("-r", "--refresh",  dest='web_interval', type=int, help="Webserver workers refresh interval, seconds",
                             default=conf_get_default("webserver", "worker_refresh_interval", 30))
    init_parser.add_argument("-w", "--workers",  dest='web_workers', type=int, help="Webserver workers refresh batch size",
                             default=conf_get_default("webserver", "worker_refresh_batch_size", 1))
    init_parser.add_argument("-p", "--threads",  dest='threads', type=int, help="Max Airflow Scheduler threads",
                             default=conf_get_default("scheduler", "max_threads", 2))


    submit_parser = subparsers.add_parser('submit', help="Submit custom workflow", parents=[parent_parser])
    submit_parser.set_defaults(func=submit_job)
    submit_parser.add_argument("-o", "--outdir", dest='output_folder', type=str, help="Output directory. Default: ./", default=".")
    submit_parser.add_argument("-t", "--tmp", dest='tmp_folder', type=str, help="Folder to store temporary data. Default: /tmp")
    submit_parser.add_argument("-u", "--uid", dest='uid', type=str, help="Experiment unique ID. Default: random uuid", default=str(uuid.uuid4()))
    submit_parser.add_argument("-r", "--run", dest='run', action="store_true", help="Run workflow with Airflow Scheduler")
    submit_parser.add_argument("workflow", type=str, help="Workflow file path")
    submit_parser.add_argument("job", type=str, help="Job file path")


    demo_parser = subparsers.add_parser('demo', help="Run demo workflows", parents=[parent_parser])
    demo_parser.set_defaults(func=run_demo)
    demo_parser.add_argument("-o", "--outdir", dest='output_folder', type=str, help="Output directory. Default: ./", default=".")
    demo_parser.add_argument("-t", "--tmp", dest='tmp_folder', type=str, help="Folder to store temporary data. Default: /tmp")
    demo_parser.add_argument("-u", "--uid", dest='uid', type=str, help="Experiment's unique ID; ignored with -a/-l arguments. Default: random uuid", default=str(uuid.uuid4()))
    demo_parser.add_argument("workflow", type=str, help="Demo workflow name from the list")


    excl_group = demo_parser.add_mutually_exclusive_group()
    excl_group.add_argument("-a", "--auto", dest='auto', action="store_true", help="Run all demo workflows with Airflow Webserver & Scheduler")
    excl_group.add_argument("-m", "--manual", dest='manual', action="store_true", help="Submit all demo workflows. Requires Airflow Webserver & Scheduler to be run separately")
    excl_group.add_argument("-l", "--list", dest='list', action="store_true", help="List demo workflows")

    return general_parser


def run_demo_auto(args):
    run_demo_manual(args)
    start_background_scheduler()
    logging.info("Run Airflow Webserver in background\n- to open webserver follow the link {}".format(get_webserver_url()))
    with Mute():
        webserver(get_airflow_default_args("webserver"))


def run_demo_manual(args):
    clean_jobs_folder()
    workflows = get_demo_workflow()
    for idx, wf in enumerate(workflows):
        logging.info("Process demo workflow {}/{}".format(idx+1, len(workflows)))
        submit_job(get_updated_args(args, wf))


def run_demo(args):
    asset_conf()
    if args.auto:
        run_demo_auto(args)
    elif args.manual:
        run_demo_manual(args)
        logging.info("To process submitted workflows run Airflow Scheduler separately")
    elif args.list:
        print("Available demo workflows:")
        for wf in get_demo_workflow():
            print("-", wf["workflow"]["name"])
    elif args.workflow:
        try:
            submit_job(get_updated_args(args, get_demo_workflow(args.workflow)[0], keep_uid=True, keep_output_folder=True))
        except IndexError:
            logging.warning("{} is not found in the demo workflows list".format(args.workflow))
        logging.info("To process submitted workflows run Airflow Scheduler separately")
    else:
        arg_parser().parse_known_args(["demo", "--help"])


def run_init(args):
    asset_conf("init")
    logging.info("Init cwl-airflow")
    update_config(args)
    create_folders()
    export_dags()
    logging.info("Init Airflow DB")
    with Mute():
        initdb(argparse.Namespace())


def submit_job(args):
    asset_conf()
    logging.info("Load workflow\n- workflow: {workflow}\n- job:      {job}\n- uid:      {uid}".format(**vars(args)))
    exit_if_unsupported_feature(args.workflow)
    export_job_file(args)
    if getattr(args, "run", None):
        logging.info("Run Airflow Scheduler")
        with Mute():
            add_run_info(args)
            scheduler(args)
        print(get_workflow_output(args.dag_id))


def main(argsl=None):
    if argsl is None:
        argsl = sys.argv[1:]
    argsl.append("")  # To avoid raising error when argsl is empty
    args, _ = arg_parser().parse_known_args(argsl)
    args = normalize_args(args, skip_list=["func", "uid", "limit", "dag_timeout", "threads",
                                           "web_interval", "web_workers", "run", "auto", "manual", "list", "quiet"])
    with Mute():
        reset_root_logger(args.quiet)
    args.func(args)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



import os
from airflow.models import DAG
from cwl_airflow.dag_components.cwlstepoperator import CWLStepOperator
from cwl_airflow.utils.utils import (shortname, flatten, load_cwl, convert_to_workflow)
from cwl_airflow.dag_components.jobdispatcher import JobDispatcher
from cwl_airflow.dag_components.jobcleanup import JobCleanup


class CWLDAG(DAG):

    @property
    def filepath(self):
        """
        Overwritten: File location of where the dag object is instantiated
        """
        return self.full_filepath

    def __init__(self, dag_id, default_args, *args, **kwargs):
        super(self.__class__, self).__init__(dag_id=dag_id, default_args=default_args, *args, **kwargs)
        self.cwlwf = load_cwl(default_args["job_data"]["content"]["workflow"])
        if self.cwlwf.tool["class"] == "CommandLineTool" or self.cwlwf.tool["class"] == "ExpressionTool":
            # workflow_file = os.path.join(default_args["tmp_folder"], os.path.basename(default_args["job_data"]["content"]["workflow"]))

            tool_dirname = os.path.dirname(default_args["job_data"]["content"]["workflow"])
            tool_filename, tool_ext = os.path.splitext(os.path.basename(default_args["job_data"]["content"]["workflow"]))
            workflow_file = os.path.join(tool_dirname, tool_filename + '_workflow' + tool_ext)

            self.cwlwf = load_cwl(convert_to_workflow(tool=self.cwlwf.tool,
                                                      tool_file=default_args["job_data"]["content"]["workflow"],
                                                      workflow_file=workflow_file))
        self.requirements = self.cwlwf.tool.get("requirements", [])

    def create(self):
        outputs = {}
        for step in self.cwlwf.steps:
            cwl_task = CWLStepOperator(cwl_step=step, dag=self)
            outputs[shortname(step.tool["id"])] = cwl_task
            for out in step.tool["outputs"]:
                outputs[shortname(out["id"])] = cwl_task
        for step in self.cwlwf.steps:
            current_task = outputs[shortname(step.tool["id"])]
            for inp in step.tool["inputs"]:
                step_input_sources = inp.get("source", '') if isinstance(inp.get("source", ''), list) else [inp.get("source", '')]
                for source in step_input_sources:
                    parent_task = outputs.get(shortname(source), None)
                    if parent_task and parent_task not in current_task.upstream_list:
                        current_task.set_upstream(parent_task)

    def assign_job_dispatcher(self, task):
        for current_task in self.tasks:
            if isinstance(current_task, JobDispatcher) or isinstance(current_task, JobCleanup):
                continue
            current_task_input_sources = [shortname(source) for source in flatten([current_task_input["source"] \
                                                                                   for current_task_input in current_task.cwl_step.tool["inputs"] \
                                                                                   if "source" in current_task_input])]
            workflow_input_id = [shortname(workflow_input["id"]) for workflow_input in self.cwlwf.tool["inputs"]]
            # Should also check if current_task is on top, 'cos if task has all parameters to be set by
            # default and don't need any of its inputs to be read from the file
            # but it suppose to either return something directly to workflow output
            # or through the other tasks which don't have connections with JobDispatcher too,
            # it may happen that it will lost 'outdir', because the last one is set only by JoDispatcher task
            if any(i in current_task_input_sources for i in workflow_input_id) or not current_task.upstream_list:
                current_task.set_upstream(task)

    def assign_job_cleanup(self, task):
        for current_task in self.tasks:
            if isinstance(current_task, JobCleanup):
                continue
            if isinstance(current_task, JobDispatcher):
                current_task_outputs_id = [shortname(current_task_output["id"]) for current_task_output in current_task.dag.cwlwf.tool["inputs"]]
            else:
                current_task_outputs_id = [shortname(current_task_output["id"]) for current_task_output in current_task.cwl_step.tool["outputs"]]
            workflow_outputs_outputsource = [shortname(workflow_output["outputSource"]) for workflow_output in self.cwlwf.tool["outputs"]]
            # print "current_task_outputs_id: \n", yaml.round_trip_dump(current_task_outputs_id)
            # print "workflow_outputs_outputsource: \n", yaml.round_trip_dump(workflow_outputs_outputsource)
            if any(i in current_task_outputs_id for i in workflow_outputs_outputsource):
                task.set_upstream(current_task)

    def get_output_list(self):
        # return [shortname(o) for o in self.cwlwf.tool["outputs"] ]
        outputs = {}
        for out in self.cwlwf.tool["outputs"]:
            outputs[shortname(out["outputSource"])] = shortname(out["id"])
        return outputs



import json
import shutil
import logging
from jsonmerge import merge
from airflow.models import BaseOperator
from cwltool.process import relocateOutputs
from cwltool.stdfsaccess import StdFsAccess


class JobCleanup(BaseOperator):

    def __init__(self, *args, **kwargs):
        super(JobCleanup, self).__init__(task_id=self.__class__.__name__, *args, **kwargs)

    def execute(self, context):
        collected_outputs = {}
        for task_outputs in self.xcom_pull(context=context, task_ids=[task.task_id for task in self.upstream_list]):
            collected_outputs = merge(collected_outputs, task_outputs["outputs"])
        logging.debug('Collected outputs:\n{}'.format(json.dumps(collected_outputs, indent=4)))
        tmp_folder = collected_outputs["tmp_folder"]
        output_folder = collected_outputs["output_folder"]
        relocated_outputs = relocateOutputs(outputObj={output_id: collected_outputs[output_src]
                                                       for output_src, output_id in self.dag.get_output_list().items()
                                                       if output_src in collected_outputs},
                                            outdir=output_folder,
                                            output_dirs=[output_folder],
                                            action="copy",
                                            fs_access=StdFsAccess(""))

        relocated_outputs = {key.split("/")[-1]: val for key, val in relocated_outputs.items()}
        shutil.rmtree(tmp_folder, ignore_errors=False)
        logging.debug('Delete temporary output directory {}'.format(tmp_folder))
        logging.info("WORKFLOW RESULTS\n" + json.dumps(relocated_outputs, indent=4))



import sys
import os
import json
import copy
import logging
import tempfile
from airflow.models import BaseOperator
from jsonmerge import merge
import cwltool.executors
import cwltool.workflow
import cwltool.errors
import cwltool.stdfsaccess
from cwltool.context import RuntimeContext, getdefault
from cwl_airflow.utils.utils import (shortname, flatten)
from airflow.utils.log.logging_mixin import StreamLogWriter


class StreamLogWriterUpdated (StreamLogWriter):

    def fileno(self):
        return -1


class CWLStepOperator(BaseOperator):

    def __init__(self, cwl_step, *args, **kwargs):
        self.cwl_step = cwl_step
        super(self.__class__, self).__init__(task_id=shortname(cwl_step.tool["id"]).split("/")[-1], *args, **kwargs)

    def execute(self, context):
        logging.info('Running tool: \n{}'.format(json.dumps(self.cwl_step.tool, indent=4)))
        collected_outputs = {}
        for task_outputs in self.xcom_pull(context=context, task_ids=[task.task_id for task in self.upstream_list]):
            collected_outputs = merge(collected_outputs, task_outputs["outputs"])
        logging.debug('Collected outputs:\n{}'.format(json.dumps(collected_outputs, indent=4)))

        jobobj = {}

        for inp in self.cwl_step.tool["inputs"]:
            jobobj_id = shortname(inp["id"]).split("/")[-1]
            source_ids = []
            promises_outputs = []
            try:
                source_ids = [shortname(source) for source in inp["source"]] if isinstance(inp["source"], list) else [shortname(inp["source"])]
                promises_outputs = [collected_outputs[source_id] for source_id in source_ids if source_id in collected_outputs]
            except Exception as ex:
                logging.info("{0}: Couldn't find source field in step input:\n{1}".format(self.task_id,json.dumps(inp,indent=4)))
            logging.info('For input {} with source_ids: {} found upstream outputs: \n{}'.format(jobobj_id, source_ids, promises_outputs))
            if len(promises_outputs) > 1:
                if inp.get("linkMerge", "merge_nested") == "merge_flattened":
                    jobobj[jobobj_id] = flatten(promises_outputs)
                else:
                    jobobj[jobobj_id] = promises_outputs
            elif len(promises_outputs) == 1 and (promises_outputs[0] is not None): # Should also check if [None], because in this case we need to take default value
                jobobj[jobobj_id] = promises_outputs[0]
            elif "valueFrom" in inp:
                jobobj[jobobj_id] = None
            elif "default" in inp:
                d = copy.copy(inp["default"])
                jobobj[jobobj_id] = d
            else:
                continue


        logging.info('{0}: Collected job object: \n {1}'.format(self.task_id, json.dumps(jobobj,indent=4)))

        valueFrom = {
            shortname(i["id"]).split("/")[-1]: i["valueFrom"] for i in self.cwl_step.tool["inputs"]
            if "valueFrom" in i}

        logging.info('{0}: Step inputs with valueFrom: \n{1}'.format(self.task_id, json.dumps(valueFrom,indent=4)))

        def postScatterEval(shortio):
            def valueFromFunc(k, v):
                if k in valueFrom:
                    return cwltool.workflow.expression.do_eval(
                        valueFrom[k], shortio, self.dag.requirements,
                        None, None, {}, context=v)
                else:
                    return v
            return {k: valueFromFunc(k, v) for k, v in shortio.items()}

        job = postScatterEval(jobobj)
        logging.info('{0}: Collected job object after valueFrom evaluation: \n {1}'.format(self.task_id, json.dumps(job,indent=4)))
        # maybe need to add here scatter functionality too

        kwargs = self.dag.default_args
        tmp_folder = collected_outputs["tmp_folder"]
        output_folder = collected_outputs["output_folder"]
        kwargs['outdir'] = tempfile.mkdtemp(dir=tmp_folder, prefix="step_tmp_")
        kwargs['tmpdir_prefix'] = os.path.join(tmp_folder, "cwl_tmp_")
        kwargs['tmp_outdir_prefix'] = os.path.join(tmp_folder, "cwl_outdir_tmp_")
        kwargs['rm_tmpdir'] = False
        kwargs["basedir"] = os.path.abspath(os.path.dirname(self.dag.default_args["job_data"]["path"]))

        logger = logging.getLogger("cwltool")
        sys.stdout = StreamLogWriterUpdated(logger, logging.INFO)
        sys.stderr = StreamLogWriterUpdated(logger, logging.WARN)

        executor = cwltool.executors.SingleJobExecutor()
        runtimeContext = RuntimeContext(kwargs)
        runtimeContext.make_fs_access = getdefault(runtimeContext.make_fs_access, cwltool.stdfsaccess.StdFsAccess)

        for inp in self.cwl_step.tool["inputs"]:
            if inp.get("not_connected"):
                del job[shortname(inp["id"].split("/")[-1])]

        (output, status) = executor(self.cwl_step.embedded_tool,
                                    job,
                                    runtimeContext,
                                    logger=logger)

        if not output and status == "permanentFail":
            raise ValueError

        logging.info(
            '{0}: Embedded tool outputs: \n {1}'.format(self.task_id, json.dumps(output,indent=4)))

        promises = {}
        for out in self.cwl_step.tool["outputs"]:
            out_id = shortname(out["id"])
            jobout_id = out_id.split("/")[-1]
            try:
                promises[out_id] = output[jobout_id]
            except:
                continue

        promises["tmp_folder"] = tmp_folder
        promises["output_folder"] = output_folder
        data = {"outputs": promises}

        logging.info(
            '{0}: Output: \n {1}'.format(self.task_id, json.dumps(data, indent=4)))

        return data



import os
import sys
import configparser
import argparse
import uuid
import logging
import shutil
import subprocess
from multiprocessing import Process
from json import dumps
from datetime import datetime
from cwl_airflow.utils.mute import Mute
from airflow import conf as conf
from airflow.models import DagRun
from airflow.utils.state import State
from airflow.settings import DAGS_FOLDER, AIRFLOW_HOME
from airflow.bin.cli import get_dag, CLIFactory, scheduler
from airflow.exceptions import AirflowConfigException
from cwl_airflow.utils.utils import (set_logger,
                                     gen_dag_id,
                                     get_folder,
                                     load_job,
                                     list_files,
                                     export_to_file,
                                     norm_path,
                                     get_files,
                                     conf_get_default)
from cwl_airflow.dag_components.cwldag import CWLDAG
from cwl_airflow.dag_components.jobdispatcher import JobDispatcher
from cwl_airflow.dag_components.jobcleanup import JobCleanup


def get_demo_workflow(target_wf=None, job_ext=".json"):
    workflows = get_files(norm_path(os.path.join(os.path.dirname(os.path.abspath(os.path.join(__file__, "../"))), "tests/cwl/workflows")))
    jobs = get_files(norm_path(os.path.join(os.path.dirname(os.path.abspath(os.path.join(__file__, "../"))), "tests/job")))
    combined_data = []
    for wf_name, wf_path in workflows.items():
        job_name = os.path.splitext(wf_name)[0] + job_ext
        if job_name in jobs:
            combined_data.append({"workflow": {"name": wf_name,
                                               "path": wf_path},
                                  "job": {"name": job_name,
                                          "path": jobs[job_name]}
                                  })
    return [item for item in combined_data if item["workflow"]["name"] == os.path.basename(target_wf)] if target_wf else combined_data


def export_job_file(args):
    job_entry = load_job(args.job)
    del job_entry["id"]
    job_entry['workflow'] = job_entry.get("workflow", args.workflow)
    job_entry['output_folder'] = job_entry.get("output_folder", args.output_folder)
    job_entry["uid"] = job_entry.get("uid", args.uid)
    tmp_folder = job_entry.get("tmp_folder", args.tmp_folder)
    if tmp_folder:
        job_entry['tmp_folder'] = tmp_folder
    root, ext = os.path.splitext(os.path.basename(args.job))
    copy_counter = 0
    while True:
        suffix = "" if copy_counter == 0 else "-" + str(copy_counter)
        output_filename = os.path.join(conf.get('cwl', 'jobs'), root + suffix + ext)
        if not os.path.exists(output_filename):
            args.job = output_filename
            export_to_file(args.job, dumps(job_entry, indent=4))
            logging.info("Save job file as\n- {}".format(args.job))
            break
        else:
            copy_counter += 1


def add_run_info(args):
    vars(args).update(vars(get_airflow_default_args("scheduler")))
    args.dag_id = gen_dag_id(os.path.join(conf.get('cwl', 'jobs'), os.path.basename(args.job)))
    args.num_runs = len(get_dag(args).tasks) + 3


def get_updated_args(args, workflow, keep_uid=False, keep_output_folder=False):
    updated_args = argparse.Namespace(**vars(args))
    updated_args.workflow = workflow["workflow"]["path"]
    updated_args.job = workflow["job"]["path"]
    if not keep_uid:
        updated_args.uid = str(uuid.uuid4())
    if not keep_output_folder:
        updated_args.output_folder = get_folder(os.path.join(args.output_folder, updated_args.uid))
    return updated_args


def get_active_jobs(jobs_folder, limit=10):
    """
    :param jobs_folder: job_folder: abs path to the folder with job json files  
    :param limit: max number of jobs to return
    :return: 
    """
    all_jobs = []
    for job_path in list_files(abs_path=jobs_folder, ext=[".json", ".yml", ".yaml"]):
        dag_id = gen_dag_id(job_path)
        dag_runs = DagRun.find(dag_id)
        all_jobs.append({"path": job_path,
                         "creation_date": datetime.fromtimestamp(os.path.getctime(job_path)),
                         "content": load_job(job_path),
                         "dag_id": dag_id,
                         "state": dag_runs[0].state if len(dag_runs) > 0 else State.NONE})
    success_jobs = sorted([j for j in all_jobs if j["state"] == State.SUCCESS], key=lambda k: k["creation_date"], reverse=True)[:limit]
    running_jobs = sorted([j for j in all_jobs if j["state"] == State.RUNNING], key=lambda k: k["creation_date"], reverse=True)[:limit]
    failed_jobs =  sorted([j for j in all_jobs if j["state"] == State.FAILED],  key=lambda k: k["creation_date"], reverse=True)[:limit]
    unknown_jobs = sorted([j for j in all_jobs if j["state"] == State.NONE],    key=lambda k: k["creation_date"], reverse=True)[:limit]
    return success_jobs + running_jobs + failed_jobs + unknown_jobs


def make_dag(job):
    """
    :param job: {"content": job_entry,
                 "path": job,
                 "creation_date": datetime.fromtimestamp(os.path.getctime(job_path)),
                 "dag_id": gen_dag_id(job_entry["workflow"], job_path)}
    :return:
    """
    set_logger()
    default_args = {
        'start_date': job["creation_date"],
        "job_data":   job
    }

    dag = CWLDAG(
        dag_id=job["dag_id"],
        schedule_interval='@once',
        default_args=default_args)
    dag.create()
    dag.assign_job_dispatcher(JobDispatcher(dag=dag))
    dag.assign_job_cleanup(JobCleanup(dag=dag))
    return dag


def update_config(args):
    logging.info("Update Airflow configuration")
    with open(conf.AIRFLOW_CONFIG, 'w') as output_stream:
        try:
            conf.conf.add_section('cwl')
        except configparser.DuplicateSectionError:
            pass
        conf.set('core', 'dags_are_paused_at_creation', 'False')
        conf.set('core', 'load_examples', 'False')
        conf.set('cwl', 'jobs', str(args.jobs))
        conf.set('cwl', 'limit', str(args.limit))
        conf.set('core', 'dagbag_import_timeout', str(args.dag_timeout))
        conf.set('scheduler', 'max_threads', str(args.threads))
        conf.set('webserver', 'worker_refresh_interval', str(args.web_interval))
        conf.set('webserver', 'worker_refresh_batch_size', str(args.web_workers))
        conf.set('webserver', 'hide_paused_dags_by_default', 'True')
        conf.conf.write(output_stream)


def export_dags():
    logging.info("Export cwl_dag.py to\n- {}".format(DAGS_FOLDER))
    dag_content = u"#!/usr/bin/env python3\nfrom airflow import DAG\nfrom cwl_airflow.create_dag import create_dags\nfor id, dag in create_dags().items():\n    globals()[id] = dag"
    export_to_file(os.path.join(DAGS_FOLDER, "cwl_dag.py"), dag_content)


def create_folders():
    logging.info("Create folders for jobs and dags\n- {}\n- {}".format(conf.get('cwl', 'jobs'), DAGS_FOLDER))
    get_folder(conf.get('cwl', 'jobs'))
    get_folder(DAGS_FOLDER)


def get_airflow_default_args(subparser):
    args, _ = CLIFactory.get_parser().parse_known_args([subparser])
    delattr(args, 'func')
    return args


def start_background_scheduler():
    logging.info("Run Airflow Scheduler in background")
    scheduler_thread = Process(target=scheduler, args=(get_airflow_default_args("scheduler"),))
    with Mute():
        scheduler_thread.start()


def clean_jobs_folder(folder=None):
    folder = folder if folder else conf_get_default('cwl', 'jobs', None)
    if folder and os.path.isdir(folder):
        logging.info("Cleaning jobs folder\n- {}".format(folder))
        for item in os.listdir(folder):
            path = os.path.join(folder, item)
            try:
                os.remove(path)
            except OSError:
                shutil.rmtree(path, ignore_errors=False)


def asset_conf(mode=None):

    def config():
        conf.get('cwl', 'jobs')
        conf.get('cwl', 'limit')

    def paths():
        items = [conf.get('cwl', 'jobs'), DAGS_FOLDER, os.path.join(DAGS_FOLDER, "cwl_dag.py")]
        for item in items:
            if not os.path.exists(item):
                raise FileExistsError(item)

    def docker():
        with open(os.devnull, 'w') as devnull:
            subprocess.run("docker -v", check=True, shell=True, stdout=devnull, stderr=devnull)

    def airflow():
        with open(os.devnull, 'w') as devnull:
            subprocess.run("airflow -h", check=True, shell=True, stdout=devnull, stderr=devnull)

    check_set = {
        "init": [docker, airflow],
        None:   [docker, airflow, paths, config]
    }

    for check_criteria in check_set[mode]:
        try:
            check_criteria()
        except AirflowConfigException as ex:
            logging.error("Missing required configuration\n- {}\n- run cwl-airflow init".format(str(ex)))
            sys.exit(0)
        except FileExistsError as ex:
            logging.error("Missing required file or directory\n- {}\n- run cwl-airflow init".format(str(ex)))
            sys.exit(0)
        except subprocess.CalledProcessError as ex:
            logging.error("Missing required installation\n- {}".format(str(ex)))
            sys.exit(0)
        except Exception as ex:
            logging.error("Unexpected exception\n- {}".format(str(ex)))
            sys.exit(1)

def get_webserver_url():
    return "{}:{}".format(conf.get('webserver', 'WEB_SERVER_HOST'), conf.get('webserver', 'WEB_SERVER_PORT'))


import os
import subprocess
import time
from setuptools import setup, find_packages


GIT_VERSION_FILE = os.path.join('cwl_airflow','git_version')
HERE = os.path.abspath(os.path.dirname(__file__))


def get_git_tag():
    return subprocess.check_output(['git', 'describe', '--contains']).strip()


def get_git_timestamp():
    gitinfo = subprocess.check_output(
        ['git', 'log', '--first-parent', '--max-count=1',
         '--format=format:%ct', '.']).strip()
    return time.strftime('%Y%m%d%H%M%S', time.gmtime(int(gitinfo)))


def get_description():
    README = os.path.join(HERE, 'README.md')
    with open(README, 'r') as f:
        return f.read()


def get_version():
    '''
    Tries to get pachage version with following order:
    0. default version
    1. from git_version file - when installing from pip, this is the only source to get version
    2. from tag
    3. from commit timestamp
    Updates/creates git_version file with the package version
    :return: package version 
    '''
    version = '1.0.0'                                      # set default version
    try:
        with open(GIT_VERSION_FILE, 'r') as input_stream:  # try to get version info from file
            version = input_stream.read()
    except Exception:
        pass
    try:
        version = get_git_tag()                            # try to get version info from the closest tag
    except Exception:
        try:
            version = '1.0.' + get_git_timestamp()         # try to get version info from commit date
        except Exception:
            pass
    try:
        with open(GIT_VERSION_FILE, 'w') as output_stream: # save updated version to file (or the same)
            output_stream.write(version)
    except Exception:
        pass
    return version


setup(
    name='cwl-airflow',
    description='Python package to extend Airflow functionality with CWL v1.0 support',
    long_description=get_description(),
    long_description_content_type="text/markdown",
    version=get_version(),
    url='https://github.com/Barski-lab/cwl-airflow',
    download_url='https://github.com/Barski-lab/cwl-airflow',
    author='Michael Kotliar',
    author_email='misha.kotliar@gmail.com',
    license = 'Apache-2.0',
    packages=find_packages(),
    package_data={'cwl_airflow': ['git_version']},
    include_package_data=True,
    install_requires=[
        'cwltool==1.0.20180622214234',
        'jsonmerge',
        "apache-airflow==1.9.0",
        "cryptography",
        "uuid"
    ],
    zip_safe=False,
    entry_points={
        'console_scripts': [
            "cwl-airflow=cwl_airflow.main:main",
            "cwl-runner=cwl_airflow.main:main"
        ]
    }
)


import logging
from airflow import settings


log_format = '[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s\n%(message)s'


def reset_root_logger(quiet=None):
    for log_handler in logging.root.handlers:
        logging.root.removeHandler(log_handler)
    for log_filter in logging.root.filters:
        logging.root.removeFilter(log_filter)
    level = logging.WARN if quiet else settings.LOGGING_LEVEL
    logging.basicConfig(level=level, format=log_format)
